/**
Class AgentSpawner:
Spawns a new agent, assigns it a digital signature and dispatches it to a remote host
*/

import java.io.*;
import java.net.*;
import java.security.*;


public class AgentSpawner
{
	String remoteHostName;
	String agentClassName;
	Socket clientSocket=null;
	ObjectOutputStream outBound=null;
	ObjectInputStream inBound=null;
	AgentIdentity agentIdentity;
	ByteCode agentByteCode;
	ResultObject resultObject;
	Agent myAgent;
	boolean connected=false;
	AgentServerGUI gui;


	// Constructor
	public AgentSpawner(String remoteHostName, String agentClassName, AgentIdentity agentIdentity, Agent agent, AgentServerGUI gui)
	{
		this.gui=gui;
		this.remoteHostName=remoteHostName;
		this.agentClassName=agentClassName;
		this.agentIdentity=agentIdentity;
		this.myAgent=agent;
	}


	public ErrorObject sendAgent()
	{
		ErrorObject message;

		// Check the error code for the connection establishment to the remote server
		message=setConnection();
		if (message.errorCode!=0)
			return message;
		connected=true;

		// Check the error code for the agent transfer
		message=dispatchAgent();
		return message;
	
	}

	private ErrorObject setConnection()
	{
		ErrorObject message=new ErrorObject();

		try
		{
			// Set up connection to the destination host
			clientSocket=new Socket(remoteHostName,5100);
			inBound=new ObjectInputStream(clientSocket.getInputStream());
			outBound=new ObjectOutputStream(clientSocket.getOutputStream());
		}
		catch (UnknownHostException e)
		{
			message.errorCode=-1;
			message.errorMessage=" Invalid address for the remote host. Cannot Establish Connection...";
			fatalErrorHandler();
			return message;
		}
		catch (IOException e)
		{
			message.errorCode=-1;
			message.errorMessage="Network Error. Cannot connect to remote host...";
			fatalErrorHandler();
			return message;
		}
		

		System.out.println("Connection to the remote host established. Beginning agent transfer...");
		message.errorCode=0;
		message.errorMessage="Connection Successfully Established...";
		return message;
	}

	public ErrorObject dispatchAgent()
	{
		boolean transactionResult=false;
		ErrorObject message=new ErrorObject();

		agentByteCode=new ByteCode(agentClassName);
		int errorCode=makeByteCodeObject();
		if (errorCode==-1)
		{
			message.errorCode=-1;
			message.errorMessage="Could not find the agent class in the current path. Please check path and class file...";
			return message;
		}

		// Sign the agent before sending it
		TestSig signature=new TestSig();
		ReturnSig returnSig=signature.generateSig(agentByteCode.contents);
		myAgent.signAgent(returnSig.pub, returnSig.digest, returnSig.encryptedDigest);

		try
		{
			// First send the agent identity object
			outBound.writeObject(agentIdentity);
		}
		catch (IOException e)
		{
			message.errorCode=-1;
			message.errorMessage="Network Error. Cannot connect to remote host...";
			fatalErrorHandler();
			return message;
		}

		try
		{
			// Receive the result object
			resultObject=(ResultObject) inBound.readObject();

			// check the result object
			transactionResult=checkResults();
			if (transactionResult==false)
			{
				message.errorCode=-1;
				message.errorMessage="Remote Host refused to accept the agent. Reason received:"+resultObject.errorMessage;
				fatalErrorHandler();
				return message;
			}

			// Send the byte code of the agent
			

			outBound.writeObject(agentByteCode);

			// Send the serialized agent
			// test
			outBound.writeObject(myAgent);

			// Receive the ResultObject again
	/*	resultObject=(ResultObject) inBound.readObject();
			transactionResult=checkResults();
			if (transactionResult==false)
			{
				message.errorCode=0;
				message.errorMessage="OK";
				gui.showErrorMessage("Remote Host refused agent. Reason received: "+resultObject.errorMessage);
				return message;
			}*/

			
		}
		catch (ClassNotFoundException e)
		{
			message.errorCode=-1;
			message.errorMessage="ERROR: Incoming class not of ResultObject type";
			fatalErrorHandler();
			return message;
		}
		catch (IOException e)
		{
			message.errorCode=-1;
			message.errorMessage="ERROR: Cannot send the agent object";
			fatalErrorHandler();
			return message;
		}

		System.out.println("Agent transfer complete");
		message.errorCode=0;
		message.errorMessage="Agent Successfully Dispatched to Remote Location";
		return message;
	}

	
	private int makeByteCodeObject()
	{
		File byteFile=null;
		RandomAccessFile rByteFile=null;

		try
		{
			byteFile=new File(agentByteCode.className);
			if (!(byteFile.exists()))
			{
				System.out.println("ERROR: "+agentByteCode.className+" class file not found");
				fatalErrorHandler();
				return -1;
			}
			rByteFile=new RandomAccessFile(byteFile,"r");
		}
		catch (IOException e)
		{
			System.out.println("ERROR: Cannot obtain file handles");
			fatalErrorHandler();
		}

		long fileLength=byteFile.length();
		int intLength=(int)fileLength;
		byte fileBuffer[]=new byte[intLength];
		try
		{
			rByteFile.read(fileBuffer,0,intLength);
			rByteFile.close();
		}
		catch (IOException e)
		{
			System.out.println("ERROR:IO Error in reading class file"+e);
			fatalErrorHandler();
			return -1;
		}

		agentByteCode.insertByteCode(fileBuffer);
		return 0;
	}

		

	private boolean checkResults()
	{
		if (resultObject.transactionStatus==true)
		{
			System.out.println("Remote Host says:"+resultObject.errorMessage);
			return true;
		}
		return false;
	}

	private void closeStreams()
	{
		try
		{
			inBound.close();
			outBound.close();
			clientSocket.close();
		}
		catch (IOException e)
		{
			System.out.println("IO Exception:"+e);
		}
	}

	private void fatalErrorHandler()
	{
		System.out.println("Aborting transaction...\n");
		if (connected)
		closeStreams();
	}


}



		