/**
Class RequestServicer:
Services incoming requests for agent transfers. Actually implements the agent transfer protocol from
the receiver side. Also is responsible for authentication and starting the agent

  */

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;

public class RequestServicer extends Thread
{
	private Socket clientSocket=null;
	private ObjectOutputStream outBound=null;
	private ObjectInputStream inBound=null;
	AgentIdentity incomingAgentIdentity;
	ResultObject resultObject;
	Agent agent;
	ByteCode agentByteCode;
	IncomingAgentMonitor monitor;
	AgentSpawner agentSpawner;
	String agentClassName;
	AgentServerGUI gui=null;
	protected AgentEventListener agentListener=null;
	boolean askBeforeDispatch;
	int waiterValue=0;


	// constructor
	public RequestServicer(Socket clientSocket, IncomingAgentMonitor monitor, AgentServerGUI gui,boolean ask)
	{
		this.askBeforeDispatch=ask;
		this.gui=gui;
		this.monitor=monitor;
		this.clientSocket=clientSocket;
		try
		{
			outBound=new ObjectOutputStream(this.clientSocket.getOutputStream());
			inBound=new ObjectInputStream(this.clientSocket.getInputStream());
		}
		catch (IOException e)
		{
			System.out.println("ERROR: Cannot obtain output streams");
		}
	}

	public void addAgentEventListener(AgentEventListener listener) throws java.util.TooManyListenersException
	{
		if (agentListener!=null)
		{
			throw new TooManyListenersException("Listener Already Registered");
		}
		agentListener=listener;
		waiterValue=5000;
	}

	public void removeAgentEventListener(AgentEventListener listener)
	{
		if (agentListener==null)
		{
			throw new IllegalArgumentException("Cannot remove agent listener");
		}
		waiterValue=1000;
		agentListener=null;
	}

	protected void fireAgentIdentityArrivedEvent()
	{
		AgentEvent agentEvent;
		if (agentListener!=null)
		{
			agentEvent=new AgentEvent(this);
			agentListener.agentIdentityArrived(agentEvent);
		}
		waiter();
	}

	protected void fireAgentAuthenticatedEvent()
	{
		AgentEvent agentEvent;
		if (agentListener!=null)
		{
		
				agentEvent=new AgentEvent(this);
				String mess1=new String(agent.getAgentOrganization());
				agentEvent.message=mess1;
				agentListener.agentAuthenticated(agentEvent);
		
		}
		waiter();
	}

	protected void fireAgentAuthenticationFailedEvent()
	{
		AgentEvent agentEvent;
		if (agentListener!=null)
		{
			agentEvent=new AgentEvent(this);
			agentListener.agentAuthenticationFailed(agentEvent);
		}
		waiter();

	}

	protected void fireAgentArrivedEvent()
	{
		AgentEvent agentEvent;
		if (agentListener!=null)
		{
			agentEvent=new AgentEvent(this);
			agentListener.agentArrived(agentEvent);
		}
		waiter();
	}

	protected void fireAgentStartedEvent()
	{
		AgentEvent agentEvent;
		if (agentListener!=null)
		{
			agentEvent=new AgentEvent(this);
			agentListener.agentStarted(agentEvent);
		}
		waiter();
	}

	protected void fireAgentStoppedEvent()
	{
		AgentEvent agentEvent;
		if (agentListener!=null)
		{
			agentEvent=new AgentEvent(this);
			agentListener.agentStopped(agentEvent);
		}
		waiter();
	}

	private void waiter()
	{
		try
		{
			sleep(waiterValue);
		}
		catch (InterruptedException e)
		{}
	}

	public void run()
	{

		// Obtain the incoming agent identity object
		try
		{
			// accept the AgentIdentity Object
			incomingAgentIdentity=(AgentIdentity) inBound.readObject();
		}
		catch (IOException e)
		{
			gui.showErrorMessage("ERROR: Cannot read AgentIdentity Object...");
			fatalErrorHandler();
		}
		catch (ClassNotFoundException e1)
		{
			gui.showErrorMessage("ERROR: Incoming class not of AgentIdentity type");
			fatalErrorHandler();
		}

		fireAgentIdentityArrivedEvent();

		// Authentication successful. Accept agent
		gui.showErrorMessage("Accepting agent from remote host...");
		acceptAgent();

	}

	private boolean authenticateAgent()
	{
		PublicKey key=agent.givePublicKey();
		byte digest[]=agent.giveMessageDigest();
		byte sign[]=agent.giveSignature();

		TestSig sig=new TestSig();
		boolean verifier=sig.verifySig(digest,sign,key, agentByteCode.contents);


		return verifier;
	}

	private void acceptAgent()
	{
		boolean recognizeAgent=false;
		String errorString;
		boolean errorCode=true;
		
		// Instantiate the ResultObject which will be sent as a reply to the agent sender
		resultObject=new ResultObject();
		resultObject.errorMessage="Welcome!";
		resultObject.transactionStatus=true;
		resultObject.errorCode=0;


		// Send the ResultObject indicating to the sender to go ahead with the agent transfer
		try
		{
			outBound.writeObject(resultObject);
		}
		catch (IOException e)
		{
			gui.showErrorMessage("ERROR: Cannot send  ResultObject to agent-sender:"+e);
			fatalErrorHandler();
		}

		// Get the incoming agent's byte code
		try
		{
			// accept the ByteCode object
			agentByteCode=(ByteCode) inBound.readObject();
		}
		catch (IOException e)
		{
			gui.showErrorMessage("ERROR: IO error while getting ByteCode object:"+e);
			fatalErrorHandler();
		}
		catch (ClassNotFoundException e1)
		{
			gui.showErrorMessage("ERROR: Incoming object not of ByteCode type");
			fatalErrorHandler();
		}

		// Save the incoming agent's byte code in the current class path
		errorCode=saveByteCode();
		if (errorCode==false)
		{
			fatalErrorHandler();
		}

		// save the agent class name for use while sending the agent from this location
		agentClassName=agentByteCode.className;

		// Get the serialized state of the Agent
		try
		{
			agent=(Agent) inBound.readObject();
		}
		catch (IOException e)
		{
			gui.showErrorMessage("ERROR: Cannot obtain incoming agent");
			fatalErrorHandler();
		}
		catch (ClassNotFoundException e1)
		{
			gui.showErrorMessage("ERROR: Incoming object not of Agent type");
			fatalErrorHandler();
		}

		// The complete agent is here
		fireAgentArrivedEvent();
		gui.showErrorMessage("Agent arrived...");

		// Authenticate the agent based the the digital certificate received
		recognizeAgent=authenticateAgent();
		if (recognizeAgent==false)
		{
			// failure of authentication or some other reason
			fireAgentAuthenticationFailedEvent();
			gui.showErrorMessage("ERROR: Agent not recognized. Aborting process...");
	//		fatalErrorHandler();
			resultObject.errorMessage="Authentication Failed.";
			resultObject.transactionStatus=false;
			resultObject.errorCode=0;
		}
		else 
		{
			resultObject.errorMessage="Authentication Succeeded";
			resultObject.transactionStatus=true;
			resultObject.errorCode=0;
		}



		// Send the final result object to the sender of the agent

	/*	try
		{
			outBound.writeObject(resultObject);
		}
		catch (IOException e)
		{
			gui.showErrorMessage("ERROR: Cannot send  ResultObject to agent-sender:"+e);
			fatalErrorHandler();
		}*/

		// NEW CODE

		fireAgentAuthenticatedEvent();
		if (gui.getAskBeforeAccepting())
		{
			if (!(gui.getAcceptAgent()))
			{
				// If the user has decided NOT to accept the agent
				gui.removeAgentName(agentClassName);
				gui.showErrorMessage("Agent is not accepted. Removing agent contents...");
				closeStreams();
				gui.setAcceptAgent(true);
				this.stop();
			}
		}

		// Add agent name to the GUI list box
		gui.addAgentName(agentClassName);

		// Get the next destination for the agent
		String nextAgentHostName=agent.getNextHostAddressForDispatch();
		if (nextAgentHostName==null)
		{
			// home coming agent
			gui.removeAgentName(agentClassName);
			gui.addResultAgentName(agentClassName, agent);
			gui.showErrorMessage("Home coming agent...");
			closeStreams();
			this.stop();
			/*gui.showErrorMessage("ERROR: Cannot obtain next host name...");
			fatalErrorHandler();*/
		}
	
		fireAgentStartedEvent();
		handleAgent();
		fireAgentStoppedEvent();

		

		if (gui.askBeforeDispatch)
		{
			createAskDialog(agentClassName);
		}
		agentSpawner=new AgentSpawner(nextAgentHostName,agentClassName, incomingAgentIdentity,agent,gui);
		ErrorObject error=agentSpawner.sendAgent();
		gui.showErrorMessage("Dispatching agent to the next destination...");
		if (error.errorCode==-1)
			gui.showErrorMessage(error.errorMessage);

		cleanUpAgentRemains();
		gui.removeAgentName(agentClassName);
		closeStreams();
	}

	private void createAskDialog(String message)
	{
		ErrorDialog mess=new ErrorDialog(gui,message,400,100,false);
	}


	// SUBJECT TO CHANGE
	private void handleAgent()
	{
		agent.run();
		// Start the agent thread
		
		
	}

	private void cleanUpAgentRemains()
	{
		// remove the class file
		System.out.println("Cleaning up remains");
	}


	private void cleanUpContents()
	{
		System.out.println("Currently clean up does nothing...");
		return;
	}

	private void fatalErrorHandler()
	{
		closeStreams();
		this.stop();
	}


	private void closeStreams()
	{
		try
		{
			inBound.close();
			outBound.close();
			clientSocket.close();
		}
		catch (IOException e)
		{
			System.out.println("ERROR: Cannot close object streams");
			monitor.decrementThreadCount();
			this.stop();
		}
	}


	private boolean saveByteCode()
	{
		File byteCodeFile=null;
		RandomAccessFile rByteCodeFile=null;

		try
		{
			byteCodeFile=new File(agentByteCode.className);
			if (byteCodeFile.exists())
			{
				System.out.println(" Class file already exists. Anyway...");
			}

			rByteCodeFile=new RandomAccessFile(byteCodeFile,"rw");
			rByteCodeFile.write(agentByteCode.contents);
			rByteCodeFile.close();
		}
		catch (IOException e)
		{
			System.out.println("ERROR: Cannot write the agent byte code file");
			return false;
		}
		return true;
	}

}












